import './style.scss'
import TinyEmitter from 'tiny-emitter'
import Popper from 'popper.js'
import { addEvent, noop, getSelectionRange, rAF, createStyleNode } from './utils'
import referenceObject from './reference'

type TFunction = () => void

export interface IOptions {
  styles?: string[]
}

export default class extends TinyEmitter {
  private el: HTMLElement
  private options?: IOptions
  private container: HTMLUnknownElement
  private iFrame: HTMLIFrameElement
  private popper: Popper
  private _stopScroll: TFunction
  private unlisteners: TFunction[]

  constructor (el: HTMLElement, options?: IOptions) {
    super()
    this.el = el
    this.options = options
    this.unlisteners = []
    this._stopScroll = noop
    this.enableListeners()
  }

  enableListeners () {
    this.unlisteners.push(
      addEvent(document, 'click', () => {
        this.hide()
      }),
      addEvent(document, 'mouseup', () => {
        window.setTimeout(() => {
          this.show()
        }, 0)
      })
    )
  }

  disableListeners () {
    this.unlisteners.forEach(unlisten => {
      unlisten()
    })
    this.unlisteners = []
  }

  private _init () {
    this._init = noop

    const container = document.createElement('hs-container')
    const arrow = document.createElement('hs-arrow')
    const iFrame = document.createElement('iframe')

    container.appendChild(arrow)
    container.appendChild(iFrame)
    document.documentElement.appendChild(container)

    const { contentDocument } = iFrame
    const { head } = contentDocument
    head.appendChild(createStyleNode('body{margin:0;background-color:#fff}', contentDocument))
    contentDocument.body.appendChild(this.el)

    const { options } = this

    if (options) {
      // 在 iFrame 中插入样式
      const { styles } = options
      if (Array.isArray(styles)) {
        styles.forEach(str => {
          head.appendChild(createStyleNode(str, contentDocument))
        })
      }
    }

    // 初始化弹窗
    const popper = new Popper(referenceObject, container, {
      placement: 'top',
      modifiers: {
        arrow: {
          element: arrow
        }
      }
    })
    popper.disableEventListeners()

    this.container = container
    this.iFrame = iFrame
    this.popper = popper
    this.emit('created')
  }

  // 在显示弹窗的期间监听全局 scroll 事件
  private _startScroll () {
    let running = false
    const unwatch = addEvent(document, 'scroll', () => {
      if (running) return
      rAF(() => {
        this.popper.update()
        running = false
      })
    }, true)

    this._stopScroll = function () {
      unwatch()
      this._stopScroll = noop
    }
  }

  resize () {
    const { el } = this
    const { style } = this.iFrame
    style.width = el.offsetWidth + 'px'
    style.height = el.offsetHeight + 'px'
  }

  show () {
    const range = getSelectionRange()
    if (range) {
      this._init()
      this.emit('beforeShow', range.toString())
      this.popper.update()
      // this.popper.enableEventListeners()
      this.resize()
      this.container.style.display = 'block'
      this._startScroll()
      return true
    } else {
      return false
    }
  }

  hide () {
    const { container } = this
    if (container) {
      // this.popper.disableEventListeners()
      container.style.display = 'none'
      this._stopScroll()
    }
  }

  destroy () {
    this.disableListeners()
    this._stopScroll()
    this.off('created')
    this.off('beforeShow')
    const { container } = this
    if (container) {
      document.documentElement.removeChild(container)
    }
  }
}
